@@ -13,6 +13,7 @@ from .base import (
     _SwifterBaseObject,
     suppress_stdout_stderr_logging,
     ERRORS_TO_HANDLE,
+    RAY_INSTALLED,
     N_REPEATS,
 )
 
@@ -23,8 +24,11 @@ DEFAULT_KWARGS = {
     "progress_bar": True,
     "progress_bar_desc": None,
     "allow_dask_on_strings": False,
+    "force_parallel": False,
 }
 
+GROUPBY_MAX_ROWS_PANDAS_DEFAULT = 5000
+
 
 def register_default_config_dataframe_accessor(dataframe_to_register, kwargs):
     """
@@ -43,6 +47,7 @@ def register_default_config_dataframe_accessor(dataframe_to_register, kwargs):
                 desc=kwargs.get("progress_bar_desc", DEFAULT_KWARGS["progress_bar_desc"]),
             )
             .allow_dask_on_strings(enable=kwargs.get("allow_dask_on_strings", DEFAULT_KWARGS["allow_dask_on_strings"]))
+            .force_parallel(enable=kwargs.get("force_parallel", DEFAULT_KWARGS["force_parallel"]))
         )
 
     dataframe_to_register.__init__ = new_init
@@ -74,6 +79,7 @@ class _SwifterObject(_SwifterBaseObject):
         progress_bar=DEFAULT_KWARGS["progress_bar"],
         progress_bar_desc=DEFAULT_KWARGS["progress_bar_desc"],
         allow_dask_on_strings=DEFAULT_KWARGS["allow_dask_on_strings"],
+        force_parallel=DEFAULT_KWARGS["force_parallel"],
     ):
         super().__init__(base_obj=pandas_obj, npartitions=npartitions)
         if self._obj.index.duplicated().any():
@@ -87,6 +93,7 @@ class _SwifterObject(_SwifterBaseObject):
         self._progress_bar = progress_bar
         self._progress_bar_desc = progress_bar_desc
         self._allow_dask_on_strings = allow_dask_on_strings
+        self._force_parallel = force_parallel
 
     def set_dask_threshold(self, dask_threshold=1):
         """
@@ -120,6 +127,15 @@ class _SwifterObject(_SwifterBaseObject):
         self._allow_dask_on_strings = enable
         return self
 
+    def force_parallel(self, enable=True):
+        """
+        Force swifter to use dask parallel processing, without attempting any
+        vectorized solution or estimating pandas apply duration to determine
+        what will be the fastest approach
+        """
+        self._force_parallel = enable
+        return self
+
     def rolling(
         self,
         window,
@@ -150,6 +166,7 @@ class _SwifterObject(_SwifterBaseObject):
             progress_bar=self._progress_bar,
             progress_bar_desc=self._progress_bar_desc,
             allow_dask_on_strings=self._allow_dask_on_strings,
+            force_parallel=self._force_parallel,
             **kwds,
         )
 
@@ -196,6 +213,7 @@ class _SwifterObject(_SwifterBaseObject):
             progress_bar=self._progress_bar,
             progress_bar_desc=self._progress_bar_desc,
             allow_dask_on_strings=self._allow_dask_on_strings,
+            force_parallel=self._force_parallel,
             **kwds,
         )
 
@@ -272,6 +290,10 @@ class SeriesAccessor(_SwifterObject):
         if not self._nrows:
             return self._obj.apply(func, convert_dtype=convert_dtype, args=args, **kwds)
 
+        # If parallel processing is forced by the user, then skip the logic and apply dask
+        if self._force_parallel:
+            return self._dask_apply(func, convert_dtype, *args, **kwds)
+
         sample = self._obj.iloc[self._SAMPLE_INDEX]
         # check if input is string or
         # if the user is overriding the string processing default
@@ -390,6 +412,10 @@ class DataFrameAccessor(_SwifterObject):
         if not self._nrows:
             return self._obj.apply(func, axis=axis, raw=raw, result_type=result_type, args=args, **kwds)
 
+        # If parallel processing is forced by the user, then skip the logic and apply dask
+        if self._force_parallel:
+            return self._dask_apply(func, axis, raw, result_type, *args, **kwds)
+
         sample = self._obj.iloc[self._SAMPLE_INDEX]
         # check if input is string
         # or if the user is overriding the string processing default
@@ -478,6 +504,10 @@ class DataFrameAccessor(_SwifterObject):
         if not self._nrows:
             return self._obj.applymap(func)
 
+        # If parallel processing is forced by the user, then skip the logic and apply dask
+        if self._force_parallel:
+            return self._dask_applymap(func)
+
         sample = self._obj.iloc[self._SAMPLE_INDEX]
         # check if input is string
         # or if the user is overriding the string processing default
@@ -511,17 +541,136 @@ class DataFrameAccessor(_SwifterObject):
 
                 return applymap_func(func)
 
+    def groupby(
+        self, by=None, axis=0, level=None, as_index=True, sort=True, group_keys=True, observed=False, dropna=True
+    ):
+        """
+        Create a swifter groupby object
+        """
+        grpby_kwargs = {
+            "level": level,
+            "as_index": as_index,
+            "sort": sort,
+            "group_keys": group_keys,
+            "observed": observed,
+            "dropna": dropna,
+        }
+        if RAY_INSTALLED:
+            return GroupBy(
+                self._obj,
+                by=[by] if isinstance(by, str) else by,
+                axis=axis,
+                progress_bar=self._progress_bar,
+                progress_bar_desc=self._progress_bar_desc,
+                **grpby_kwargs,
+            )
+        else:
+            raise NotImplementedError(
+                "Ray is required for groupby apply functionality."
+                "Please install `ray` before continuing and then restart your script or kernel."
+            )
+
+
+if RAY_INSTALLED:  # noqa: C901
+
+    class GroupBy(DataFrameAccessor):
+        import ray
+
+        def __init__(
+            self,
+            pandas_obj,
+            by,
+            axis=0,
+            npartitions=DEFAULT_KWARGS["npartitions"],
+            dask_threshold=DEFAULT_KWARGS["dask_threshold"],
+            progress_bar=DEFAULT_KWARGS["progress_bar"],
+            progress_bar_desc="Ray GroupBy Apply",
+            **grpby_kwargs,
+        ):
+            super(GroupBy, self).__init__(
+                pandas_obj,
+                npartitions=npartitions,
+                dask_threshold=dask_threshold,
+                progress_bar=progress_bar,
+                progress_bar_desc=progress_bar_desc,
+            )
+            self._sample_pd = pandas_obj.iloc[self._SAMPLE_INDEX]
+            self._obj_pd = pandas_obj
+            self._nrows = pandas_obj.shape[0]
+            self._by = by
+            self._axis = axis
+            self._grpby_kwargs = grpby_kwargs
+
+        def _wrapped_apply(self, func, *args, **kwds):
+            def wrapped():
+                with suppress_stdout_stderr_logging():
+                    self._sample_pd.groupby(by=self._by, axis=self._axis, **self._grpby_kwargs).apply(
+                        func, *args, **kwds
+                    )
+
+            return wrapped
+
+        # NOTE: All credit for the _ray_apply/_ray_apply_chunk logic goes to github user @diditforlulz273
+        # NOTE: He provided a gist which I adapted to work in swifter's codebase
+        # NOTE: https://gist.github.com/diditforlulz273/06ffa5f5b1c00830671ce0330851352f
+        @ray.remote
+        def _ray_apply_chunk(self, chunk, func, *args, **kwds):
+            # Ray makes data immutable when stored in its memory.
+            # This approach prevents state sharing among processes, but we have a separate chunk for each process
+            # to get rid of copying data, we make it mutable in-place again by this hack
+            for d in range(len(chunk._data.blocks)):
+                try:
+                    chunk._data.blocks[d].values.flags.writeable = True
+                except Exception:
+                    pass
+
+            return chunk.groupby(self._by, axis=self._axis, **self._grpby_kwargs).apply(func, *args, **kwds)
+
+        def _ray_apply(self, func, *args, **kwds):
+            import ray
+
+            unique_groups = self._obj_pd[self._by[0]].unique()
+            n_splits = min(len(unique_groups), self._npartitions)
+            splits = np.array_split(unique_groups, n_splits)
+            chunks = [self._obj_pd.loc[self._obj_pd[self._by[0]].isin(splits[x])] for x in range(n_splits)]
+
+            # Fire and forget
+            chunk_id = [ray.put(ch) for ch in chunks]
+            ray_obj_refs = [
+                self._ray_apply_chunk.remote(self, chunk_id[i], func, *args, **kwds) for i in range(n_splits)
+            ]
+
+            # TQDM progress bar
+            if self._progress_bar:
+                for chunk in tqdm(range(n_splits), desc=self._progress_bar_desc):
+                    ray.wait(ray_obj_refs, num_returns=chunk + 1)
+
+            # Collect, sort, and return
+            return pd.concat(ray.get(ray_obj_refs), axis=self._axis).sort_index()
+
+        def apply(self, func, *args, **kwds):
+            """
+            Apply the function to the groupby swifter object
+            """
+            # if the transformed dataframe is empty or very small, return early using Pandas
+            if not self._nrows or self._nrows <= GROUPBY_MAX_ROWS_PANDAS_DEFAULT:
+                return self._obj_pd.groupby(self._by, axis=self._axis, **self._grpby_kwargs).apply(func, *args, **kwds)
+
+            # Swifter logic can't accurately estimate groupby applies, so always parallelize
+            return self._ray_apply(func, *args, **kwds)
+
 
 class Transformation(_SwifterObject):
     def __init__(
         self,
         pandas_obj,
-        npartitions=None,
-        dask_threshold=1,
-        scheduler="processes",
-        progress_bar=True,
-        progress_bar_desc=None,
-        allow_dask_on_strings=False,
+        npartitions=DEFAULT_KWARGS["npartitions"],
+        dask_threshold=DEFAULT_KWARGS["dask_threshold"],
+        scheduler=DEFAULT_KWARGS["scheduler"],
+        progress_bar=DEFAULT_KWARGS["progress_bar"],
+        progress_bar_desc=DEFAULT_KWARGS["progress_bar_desc"],
+        allow_dask_on_strings=DEFAULT_KWARGS["allow_dask_on_strings"],
+        force_parallel=DEFAULT_KWARGS["force_parallel"],
     ):
         super(Transformation, self).__init__(
             pandas_obj,
@@ -531,6 +680,7 @@ class Transformation(_SwifterObject):
             progress_bar,
             progress_bar_desc,
             allow_dask_on_strings,
+            force_parallel,
         )
         self._sample_pd = pandas_obj.iloc[: self._SAMPLE_SIZE]
         self._obj_pd = pandas_obj
@@ -556,6 +706,10 @@ class Transformation(_SwifterObject):
         if not self._nrows:
             return self._obj_pd.apply(func, *args, **kwds)
 
+        # If parallel processing is forced by the user, then skip the logic and apply dask
+        if self._force_parallel:
+            return self._dask_apply(func, *args, **kwds)
+
         # estimate time to pandas apply
         wrapped = self._wrapped_apply(func, *args, **kwds)
         timed = timeit.timeit(wrapped, number=N_REPEATS)
@@ -584,6 +738,7 @@ class Rolling(Transformation):
         progress_bar=True,
         progress_bar_desc=None,
         allow_dask_on_strings=False,
+        force_parallel=False,
         **kwds,
     ):
         super(Rolling, self).__init__(
@@ -594,6 +749,7 @@ class Rolling(Transformation):
             progress_bar=progress_bar,
             progress_bar_desc=progress_bar_desc,
             allow_dask_on_strings=allow_dask_on_strings,
+            force_parallel=force_parallel,
         )
         self._rolling_kwds = kwds.copy()
         self._comparison_pd = self._obj_pd.iloc[: self._SAMPLE_SIZE]
@@ -638,6 +794,7 @@ class Resampler(Transformation):
         progress_bar=True,
         progress_bar_desc=None,
         allow_dask_on_strings=False,
+        force_parallel=False,
         **kwds,
     ):
         super(Resampler, self).__init__(
@@ -648,6 +805,7 @@ class Resampler(Transformation):
             progress_bar=progress_bar,
             progress_bar_desc=progress_bar_desc,
             allow_dask_on_strings=allow_dask_on_strings,
+            force_parallel=force_parallel,
         )
         self._resampler_kwds = kwds.copy()
         self._comparison_pd = self._obj_pd.iloc[: self._SAMPLE_SIZE]
