@@ -40,6 +40,10 @@ def math_agg_foo(row):
     return row.sum() - row.min()
 
 
+def numeric_func(x):
+    return x["x"].mean() / x["y"].var()
+
+
 def text_foo(row):
     if row["letter"] == "A":
         return row["value"] * 3
@@ -97,6 +101,9 @@ class TestSwifter(unittest.TestCase):
         """
         Imports modin before swifter so that we have access to modin functionality
         """
+        import os
+
+        os.environ["MODIN_ENGINE"] = "dask"
         import modin.pandas as md
         import swifter
 
@@ -123,6 +130,7 @@ class TestSetup(TestSwifter):
         expected_progress_bar = False
         expected_progress_bar_desc = "TEST"
         expected_allow_dask_on_strings = True
+        expected_force_parallel = True
         set_defaults(
             npartitions=expected_npartitions,
             dask_threshold=expected_dask_threshold,
@@ -130,6 +138,7 @@ class TestSetup(TestSwifter):
             progress_bar=expected_progress_bar,
             progress_bar_desc=expected_progress_bar_desc,
             allow_dask_on_strings=expected_allow_dask_on_strings,
+            force_parallel=expected_force_parallel,
         )
         for swifter_df in [
             pd.DataFrame().swifter,
@@ -149,6 +158,7 @@ class TestSetup(TestSwifter):
             swifter_df._progress_bar == expected_progress_bar
             swifter_df._progress_bar_desc == expected_progress_bar_desc
             swifter_df._allow_dask_on_strings == expected_allow_dask_on_strings
+            swifter_df._force_parallel == expected_force_parallel
 
     def test_override_defaults(self):
         LOG.info("test_set_defaults")
@@ -160,6 +170,7 @@ class TestSetup(TestSwifter):
         set_progress_bar = False
         set_progress_bar_desc = "TEST"
         set_allow_dask_on_strings = True
+        set_force_parallel = True
 
         expected_npartitions = 3
         expected_dask_threshold = 4.5
@@ -167,6 +178,7 @@ class TestSetup(TestSwifter):
         expected_progress_bar = True
         expected_progress_bar_desc = "TEST-AGAIN"
         expected_allow_dask_on_strings = False
+        expected_force_parallel = False
         set_defaults(
             npartitions=set_npartitions,
             dask_threshold=set_dask_threshold,
@@ -174,6 +186,7 @@ class TestSetup(TestSwifter):
             progress_bar=set_progress_bar,
             progress_bar_desc=set_progress_bar_desc,
             allow_dask_on_strings=set_allow_dask_on_strings,
+            force_parallel=set_force_parallel,
         )
         for swifter_df_1, swifter_df_2 in [
             [pd.DataFrame().swifter, pd.Series().swifter],
@@ -208,6 +221,7 @@ class TestSetup(TestSwifter):
                 .set_dask_scheduler(scheduler=expected_scheduler)
                 .progress_bar(enable=expected_progress_bar, desc=expected_progress_bar_desc)
                 .allow_dask_on_strings(enable=expected_allow_dask_on_strings)
+                .force_parallel(enable=expected_force_parallel)
             )
 
             swifter_df_1._npartitions == expected_npartitions
@@ -216,6 +230,7 @@ class TestSetup(TestSwifter):
             swifter_df_1._progress_bar == expected_progress_bar
             swifter_df_1._progress_bar_desc == expected_progress_bar_desc
             swifter_df_1._allow_dask_on_strings == expected_allow_dask_on_strings
+            swifter_df_1._force_parallel == expected_force_parallel
 
             swifter_df_2._npartitions == set_npartitions
             swifter_df_2._dask_threshold == set_dask_threshold
@@ -223,6 +238,7 @@ class TestSetup(TestSwifter):
             swifter_df_2._progress_bar == set_progress_bar
             swifter_df_2._progress_bar_desc == set_progress_bar_desc
             swifter_df_2._allow_dask_on_strings == set_allow_dask_on_strings
+            swifter_df_2._force_parallel = set_force_parallel
 
     def test_set_npartitions(self):
         LOG.info("test_set_npartitions")
@@ -322,6 +338,16 @@ class TestSetup(TestSwifter):
         self.assertEqual(actual, expected)
         self.assertNotEqual(before, actual)
 
+    def test_force_parallel(self):
+        LOG.info("test_force_parallel")
+        expected = True
+        swifter_df = pd.DataFrame().swifter
+        before = swifter_df._force_parallel
+        swifter_df.force_parallel(expected)
+        actual = swifter_df._force_parallel
+        self.assertEqual(actual, expected)
+        self.assertNotEqual(before, actual)
+
     def test_stdout_redirected(self):
         LOG.info("test_stdout_redirected")
         print_messages = subprocess.check_output(
@@ -365,7 +391,7 @@ class TestPandasSeries(TestSwifter):
 
     def test_vectorized_math_apply_on_large_series(self):
         LOG.info("test_vectorized_math_apply_on_large_series")
-        df = pd.DataFrame({"x": np.random.normal(size=2_000_000)})
+        df = pd.DataFrame({"x": np.random.normal(size=10_000_000)})
         series = df["x"]
 
         tqdm.pandas(desc="Pandas Vec math apply ~ Series")
@@ -376,9 +402,7 @@ class TestPandasSeries(TestSwifter):
 
         start_swifter = time.time()
         swifter_val = (
-            series.swifter.set_npartitions(4)
-            .progress_bar(desc="Vec math apply ~ Series")
-            .apply(math_vec_square, axis=0)
+            series.swifter.set_npartitions(4).progress_bar(desc="Vec math apply ~ Series").apply(math_vec_square)
         )
         end_swifter = time.time()
         swifter_time = end_swifter - start_swifter
@@ -411,6 +435,31 @@ class TestPandasSeries(TestSwifter):
         if self.ncores > 1:  # speed test
             self.assertLess(swifter_time, pd_time)
 
+    def test_vectorized_force_parallel_math_apply_on_large_series(self):
+        LOG.info("test_vectorized_force_parallel_math_apply_on_large_series")
+        df = pd.DataFrame({"x": np.random.normal(size=2_000_000)})
+        series = df["x"]
+
+        tqdm.pandas(desc="Pandas Vec math apply ~ Series")
+        start_pd = time.time()
+        pd_val = series.progress_apply(math_vec_square)
+        end_pd = time.time()
+        pd_time = end_pd - start_pd
+
+        start_swifter = time.time()
+        swifter_val = (
+            series.swifter.set_npartitions(4)
+            .force_parallel(True)
+            .progress_bar(desc="Force Parallel - Vec math apply ~ Series")
+            .apply(math_vec_square)
+        )
+        end_swifter = time.time()
+        swifter_time = end_swifter - start_swifter
+
+        self.assertEqual(pd_val, swifter_val)  # equality test
+        if self.ncores > 1:  # speed test
+            self.assertLess(swifter_time, pd_time)
+
 
 class TestPandasDataFrame(TestSwifter):
     def test_apply_on_empty_dataframe(self):
@@ -427,6 +476,13 @@ class TestPandasDataFrame(TestSwifter):
         swifter_val = df.swifter.applymap(math_vec_square)
         self.assertEqual(pd_val, swifter_val)  # equality test
 
+    def test_groupby_apply_on_empty_dataframe(self):
+        LOG.info("test_groupby_apply_on_empty_dataframe")
+        df = pd.DataFrame(columns=["x", "y"])
+        pd_val = df.groupby("x").apply(math_vec_square)
+        swifter_val = df.swifter.groupby("x").apply(math_vec_square)
+        self.assertEqual(pd_val, swifter_val)  # equality test
+
     def test_nonvectorized_math_apply_on_small_dataframe(self):
         LOG.info("test_nonvectorized_math_apply_on_small_dataframe")
         df = pd.DataFrame({"x": np.random.normal(size=1000), "y": np.random.uniform(size=1000)})
@@ -543,6 +599,35 @@ class TestPandasDataFrame(TestSwifter):
         if self.ncores > 1:  # speed test
             self.assertLess(swifter_time, pd_time)
 
+    def test_vectorized_force_parallel_math_apply_on_large_dataframe(self):
+        LOG.info("test_vectorized_force_parallel_math_apply_on_large_dataframe")
+        df = pd.DataFrame(
+            {
+                "x": np.random.normal(size=1_000_000),
+                "y": np.random.uniform(size=1_000_000),
+            }
+        )
+
+        tqdm.pandas(desc="Pandas Nonvec math apply ~ DF")
+        start_pd = time.time()
+        pd_val = df.progress_apply(math_vec_multiply, axis=1)
+        end_pd = time.time()
+        pd_time = end_pd - start_pd
+
+        start_swifter = time.time()
+        swifter_val = (
+            df.swifter.set_npartitions(4)
+            .force_parallel(True)
+            .progress_bar(desc="Forced Parallel - Vec math apply ~ DF")
+            .apply(math_vec_multiply, axis=1)
+        )
+        end_swifter = time.time()
+        swifter_time = end_swifter - start_swifter
+
+        self.assertEqual(pd_val, swifter_val)  # equality test
+        if self.ncores > 1:  # speed test
+            self.assertLess(swifter_time, pd_time)
+
     def test_vectorized_math_applymap_on_large_dataframe(self):
         LOG.info("test_vectorized_math_applymap_on_large_dataframe")
         df = pd.DataFrame(
@@ -569,6 +654,35 @@ class TestPandasDataFrame(TestSwifter):
         if self.ncores > 1:  # speed test
             self.assertLess(swifter_time, pd_time)
 
+    def test_vectorized_force_parallel_math_applymap_on_large_dataframe(self):
+        LOG.info("test_vectorized_force_parallel_math_applymap_on_large_dataframe")
+        df = pd.DataFrame(
+            {
+                "x": np.random.normal(size=2_000_000),
+                "y": np.random.uniform(size=2_000_000),
+            }
+        )
+
+        tqdm.pandas(desc="Pandas Vec math applymap ~ DF")
+        start_pd = time.time()
+        pd_val = df.progress_applymap(math_vec_square)
+        end_pd = time.time()
+        pd_time = end_pd - start_pd
+
+        start_swifter = time.time()
+        swifter_val = (
+            df.swifter.set_npartitions(4)
+            .force_parallel(True)
+            .progress_bar(desc="Force Parallel ~ Vec math applymap ~ DF")
+            .applymap(math_vec_square)
+        )
+        end_swifter = time.time()
+        swifter_time = end_swifter - start_swifter
+
+        self.assertEqual(pd_val, swifter_val)  # equality test
+        if self.ncores > 1:  # speed test
+            self.assertLess(swifter_time, pd_time)
+
     def test_nonvectorized_math_applymap_on_large_dataframe(self):
         LOG.info("test_nonvectorized_math_applymap_on_large_dataframe")
         df = pd.DataFrame(
@@ -607,6 +721,100 @@ class TestPandasDataFrame(TestSwifter):
         swifter_val = df.swifter.progress_bar(enable=False).applymap(math_foo)
         self.assertEqual(pd_val, swifter_val)  # equality test
 
+    def test_vectorized_math_groupby_apply_on_small_dataframe(self):
+        LOG.info("test_vectorized_math_groupby_apply_on_small_dataframe")
+        df = pd.DataFrame(
+            {
+                "g": np.random.choice([0, 1, 2], size=500),
+                "x": np.random.normal(size=500),
+                "y": np.random.uniform(size=500),
+            }
+        )
+        pd_val = df.groupby("g").apply(numeric_func)
+        swifter_val = df.swifter.groupby("g").apply(numeric_func)
+        self.assertSeriesEqual(pd_val, swifter_val, "Swifter output does not equal Pandas output")  # equality test
+
+    def test_vectorized_force_parallel_math_groupby_apply_on_small_dataframe(self):
+        LOG.info("test_vectorized_force_parallel_math_groupby_apply_on_small_dataframe")
+        df = pd.DataFrame(
+            {
+                "g": np.random.choice([0, 1, 2], size=500),
+                "x": np.random.normal(size=500),
+                "y": np.random.uniform(size=500),
+            }
+        )
+        pd_val = df.groupby("g").apply(numeric_func)
+        swifter_val = df.swifter.force_parallel(True).groupby("g").apply(numeric_func)
+        self.assertSeriesEqual(pd_val, swifter_val, "Swifter output does not equal Pandas output")  # equality test
+
+    def test_vectorized_math_groupby_apply_on_large_dataframe(self):
+        LOG.info("test_vectorized_math_groupby_apply_on_large_dataframe")
+        df = pd.DataFrame(
+            {
+                "g": np.random.choice(np.arange(50000), size=500000),
+                "x": np.random.normal(size=500000),
+                "y": np.random.uniform(size=500000),
+            }
+        )
+        pd_val = df.groupby("g").apply(numeric_func)
+        swifter_val = df.swifter.groupby("g").apply(numeric_func)
+        self.assertSeriesEqual(pd_val, swifter_val, "Swifter output does not equal Pandas output")  # equality test
+
+    def test_vectorized_force_parallel_math_groupby_apply_on_large_dataframe(self):
+        LOG.info("test_vectorized_force_parallel_math_groupby_apply_on_large_dataframe")
+        df = pd.DataFrame(
+            {
+                "g": np.random.choice(np.arange(50000), size=500000),
+                "x": np.random.normal(size=500000),
+                "y": np.random.uniform(size=500000),
+            }
+        )
+        pd_val = df.groupby("g").apply(numeric_func)
+        swifter_val = df.swifter.force_parallel(True).groupby("g").apply(numeric_func)
+        self.assertSeriesEqual(pd_val, swifter_val, "Swifter output does not equal Pandas output")  # equality test
+
+    def test_vectorized_text_groupby_apply_on_small_dataframe(self):
+        LOG.info("test_vectorized_text_groupby_apply_on_small_dataframe")
+        df = pd.DataFrame(
+            {"g": np.random.choice([0, 1, 2], size=500), "text": np.random.choice(["A", "B", "C"], size=500)}
+        )
+        pd_val = df.groupby("g").apply(clean_text_foo)
+        swifter_val = df.swifter.groupby("g").apply(clean_text_foo)
+        self.assertSeriesEqual(pd_val, swifter_val, "Swifter output does not equal Pandas output")  # equality test
+
+    def test_vectorized_force_parallel_text_groupby_apply_on_small_dataframe(self):
+        LOG.info("test_vectorized_force_parallel_text_groupby_apply_on_small_dataframe")
+        df = pd.DataFrame(
+            {"g": np.random.choice([0, 1, 2], size=500), "text": np.random.choice(["A", "B", "C"], size=500)}
+        )
+        pd_val = df.groupby("g").apply(clean_text_foo)
+        swifter_val = df.swifter.force_parallel(True).groupby("g").apply(clean_text_foo)
+        self.assertSeriesEqual(pd_val, swifter_val, "Swifter output does not equal Pandas output")  # equality test
+
+    def test_vectorized_text_groupby_apply_on_large_dataframe(self):
+        LOG.info("test_vectorized_text_groupby_apply_on_large_dataframe")
+        df = pd.DataFrame(
+            {
+                "g": np.random.choice(np.arange(50000), size=500000),
+                "text": np.random.choice(["A", "B", "C"], size=500000),
+            }
+        )
+        pd_val = df.groupby("g").apply(clean_text_foo)
+        swifter_val = df.swifter.groupby("g").apply(clean_text_foo)
+        self.assertSeriesEqual(pd_val, swifter_val, "Swifter output does not equal Pandas output")  # equality test
+
+    def test_vectorized_force_parallel_text_groupby_apply_on_large_dataframe(self):
+        LOG.info("test_vectorized_force_parallel_text_groupby_apply_on_large_dataframe")
+        df = pd.DataFrame(
+            {
+                "g": np.random.choice(np.arange(50000), size=500000),
+                "text": np.random.choice(["A", "B", "C"], size=500000),
+            }
+        )
+        pd_val = df.groupby("g").apply(clean_text_foo)
+        swifter_val = df.swifter.force_parallel(True).groupby("g").apply(clean_text_foo)
+        self.assertSeriesEqual(pd_val, swifter_val, "Swifter output does not equal Pandas output")  # equality test
+
 
 class TestPandasTransformation(TestSwifter):
     def test_rolling_apply_on_empty_dataframe(self):
@@ -691,6 +899,22 @@ class TestPandasTransformation(TestSwifter):
         if self.ncores > 1:  # speed test
             self.assertLess(swifter_time, pd_time)
 
+    def test_vectorized_force_parallel_math_apply_on_large_rolling_dataframe(self):
+        LOG.info("test_vectorized_force_parallel_math_apply_on_large_rolling_dataframe")
+        df = pd.DataFrame(
+            {"x": np.arange(0, 1_000_000)},
+            index=pd.date_range("2019-01-1", "2020-01-1", periods=1_000_000),
+        )
+        pd_val = df.rolling("1d").apply(max, raw=True)
+        swifter_val = (
+            df.swifter.set_npartitions(4)
+            .force_parallel(True)
+            .rolling("1d")
+            .progress_bar(desc="Force Parallel ~ Vec math apply ~ Rolling DF")
+            .apply(max, raw=True)
+        )
+        self.assertEqual(pd_val, swifter_val)  # equality test
+
     def test_nonvectorized_math_apply_on_small_resampler_dataframe(self):
         LOG.info("test_nonvectorized_math_apply_on_small_resampler_dataframe")
         df = pd.DataFrame(
@@ -732,6 +956,33 @@ class TestPandasTransformation(TestSwifter):
         if self.ncores > 1:  # speed test
             self.assertLess(swifter_time, pd_time)
 
+    def test_nonvectorized_force_parallel_math_apply_on_large_resampler_dataframe(self):
+        LOG.info("test_nonvectorized_force_parallel_math_apply_on_large_resampler_dataframe")
+        df = pd.DataFrame(
+            {"x": np.arange(0, 1_000_000)},
+            index=pd.date_range("2019-01-1", "2020-01-1", periods=1_000_000),
+        )
+
+        start_pd = time.time()
+        pd_val = df.resample("3T").apply(math_agg_foo)
+        end_pd = time.time()
+        pd_time = end_pd - start_pd
+
+        start_swifter = time.time()
+        swifter_val = (
+            df.swifter.set_npartitions(4)
+            .force_parallel(True)
+            .resample("3T")
+            .progress_bar(desc="Force Parallel ~ Nonvec math apply ~ Resample DF")
+            .apply(math_agg_foo)
+        )
+        end_swifter = time.time()
+        swifter_time = end_swifter - start_swifter
+
+        self.assertEqual(pd_val, swifter_val)  # equality test
+        if self.ncores > 1:  # speed test
+            self.assertLess(swifter_time, pd_time)
+
 
 @run_if_modin_installed
 class TestModinSeries(TestSwifter):
@@ -759,6 +1010,11 @@ class TestModinSeries(TestSwifter):
             series.swifter.progress_bar(False)
             self.assertEqual(len(w), 1)
 
+        with warnings.catch_warnings(record=True) as w:
+            warnings.simplefilter("always")
+            series.swifter.force_parallel(False)
+            self.assertEqual(len(w), 1)
+
     def test_modin_series_errors_on_missing_transformations(self):
         LOG.info("test_modin_series_errors_on_missing_transformations")
         md = self.modinSetUp()
@@ -788,24 +1044,18 @@ class TestModinSeries(TestSwifter):
     def test_vectorized_modin_apply_on_large_series(self):
         LOG.info("test_vectorized_modin_apply_on_large_series")
         md = self.modinSetUp()
-        df = md.Series(np.random.uniform(size=20_000_000), name="x")
-        start_md = time.time()
+        df = md.Series(np.random.uniform(size=10_000_000), name="x")
+
         md_val = df.apply(math_vec_square, axis=0)
         md_pd_val = md_val._to_pandas()  # We have to bring it into pandas to confirm swifter apply speed is quicker
-        end_md = time.time()
-        md_time = end_md - start_md
 
-        start_swifter = time.time()
         swifter_val = df.swifter.set_npartitions(4).apply(math_vec_square)
         swifter_pd_val = (
             swifter_val._to_pandas()
         )  # We have to bring it into pandas to confirm swifter apply speed is quicker
-        end_swifter = time.time()
-        swifter_time = end_swifter - start_swifter
 
         self.assertEqual(md_val, swifter_val)  # equality test
         self.assertEqual(md_pd_val, swifter_pd_val)  # equality test after converting to pandas
-        self.assertLess(swifter_time, md_time)  # speed test
 
 
 @run_if_modin_installed
@@ -834,6 +1084,11 @@ class TestModinDataFrame(TestSwifter):
             df.swifter.progress_bar(False)
             self.assertEqual(len(w), 1)
 
+        with warnings.catch_warnings(record=True) as w:
+            warnings.simplefilter("always")
+            df.swifter.force_parallel(False)
+            self.assertEqual(len(w), 1)
+
     def test_modin_dataframe_errors_on_missing_transformations(self):
         LOG.info("test_modin_dataframe_errors_on_missing_transformations")
         md = self.modinSetUp()
